use std::{
    any::TypeId,
    collections::HashMap,
    error::Error,
    sync::{Arc, RwLock},
};

use vulkano::{device::Device, format::Format, pipeline::GraphicsPipeline};

use super::{GraphicsPipelineLoadFn, LoadableGraphicsPipeline};

#[derive(PartialEq, Eq)]
pub enum PipelineState {
    NeedBuild,
    Loaded,
}

pub struct PipelineLoader {
    device: Arc<Device>,
    swapchain_image_format: Format,
    depth_image_format: Format,

    // Arc<GraphicsPipeline> is used in internal of vulkano. It's not possible to use Arc<RwLock<Option<GraphicsPipeline>>> directly.
    pipelines_index: HashMap<TypeId, usize>,
    pipelines_id: Vec<TypeId>,
    pipelines_load_fn: Vec<GraphicsPipelineLoadFn>,
    // Only content is protected by Arc and RwLock to avoid push in pipeline_loader in multiple threads and just allow to lock each pipeline when is needed as parallel pipelines loading.
    // But only the pipeline loader is allowed to load a pipeline when it's needed.
    pipelines: Vec<Arc<RwLock<Option<Arc<GraphicsPipeline>>>>>,
    pipelines_state: Vec<Arc<RwLock<PipelineState>>>,
    pipelines_name: Vec<&'static str>,
}

impl PipelineLoader {
    pub fn new(
        device: Arc<Device>,
        swapchain_image_format: Format,
        depth_image_format: Format,
    ) -> Self {
        Self {
            device,
            swapchain_image_format,
            depth_image_format,
            pipelines: Vec::new(),
            pipelines_load_fn: Vec::new(),
            pipelines_name: Vec::new(),
            pipelines_id: Vec::new(),
            pipelines_state: Vec::new(),
            pipelines_index: HashMap::new(),
        }
    }

    pub fn register<T: LoadableGraphicsPipeline + 'static>(
        &mut self,
    ) -> Result<(), Box<dyn Error>> {
        let id = TypeId::of::<T>();
        self.pipelines_index.insert(id, self.pipelines.len());
        self.pipelines_id.push(id);
        self.pipelines_load_fn.push(T::load);
        self.pipelines_name.push(T::pipeline_name());
        self.pipelines_state
            .push(Arc::new(RwLock::new(PipelineState::NeedBuild)));
        self.pipelines.push(Arc::new(RwLock::new(None)));
        Ok(())
    }

    pub fn load_pipelines(&self) -> Result<(), Box<dyn Error>> {
        let iter = self
            .pipelines_name
            .iter()
            .zip(self.pipelines.iter())
            .zip(self.pipelines_load_fn.iter())
            .zip(self.pipelines_state.iter())
            .filter(|(_, state)| {
                let state = state.read().unwrap();
                *state == PipelineState::NeedBuild
            });

        for (((name, pipeline), load_fn), state) in iter {
            let new_pipeline = load_fn(
                &self.device,
                self.swapchain_image_format,
                self.depth_image_format,
            )?;
            let mut pipeline = pipeline.write().unwrap();
            *pipeline = Some(new_pipeline);
            let mut state = state.write().unwrap();
            *state = PipelineState::Loaded;
            tracing::trace!("Pipeline {name} loaded");
        }
        Ok(())
    }

    fn mark_pipelines_as_need_build(&mut self) {
        for state in self.pipelines_state.iter() {
            let mut state = state.write().unwrap();
            *state = PipelineState::NeedBuild;
        }
    }

    pub fn with_pipeline<T: 'static, F>(&self, f: F) -> Result<(), Box<dyn Error>>
    where
        F: FnOnce(&Arc<GraphicsPipeline>) -> Result<(), Box<dyn Error>>,
    {
        let id = TypeId::of::<T>();
        let index = self.pipelines_index.get(&id).ok_or("Pipeline not found")?;
        let pipeline_locker = self.pipelines[*index]
            .read()
            .map_err(|_| "Failed to lock pipeline")?;
        let pipeline = pipeline_locker.as_ref().ok_or("Pipeline not loaded")?;
        f(pipeline)
    }
}
